use axum::extract::Path;
use axum::http::HeaderMap;
use axum::Form;
use axum::{extract::State, response::IntoResponse};
use log::info;
use serde_json::Value;

use super::context::DepotContext;
use super::state::DepotState;

pub fn base_template(headers: &HeaderMap) -> Option<String> {
    let hx_request = headers.get("HX-Request").is_some();
    if hx_request {
        Some("depot/base_hx.jinja".to_string())
    } else {
        None
    }
}

pub async fn index<S: DepotState + Clone + Send + Sync + 'static>(
    depot: State<S>,
    headers: HeaderMap,
) -> impl IntoResponse {
    let templates = depot.get_templates();
    let registry = depot.get_registry();
    templates.render_html(
        "depot/index.html",
        DepotContext {
            base: base_template(&headers),
            sections: registry.get_sections(),
            ..Default::default()
        },
    )
}

// Index Action is POST to the index site. We can anchor some general business code here.
pub async fn index_action<S: DepotState + Clone + Send + Sync + 'static>(
    depot: State<S>,
) -> impl IntoResponse {
    "There is your answer!".to_owned()
}

pub async fn list_section<S: DepotState + Clone + Send + Sync + 'static>(
    depot: State<S>,
    Path(depot_key): Path<String>,
) -> impl IntoResponse {
    let templates = depot.get_templates();
    templates.render_html("depot/depot.jinja", ())
}

// List Items renders the entire list item page.
pub async fn list_item_collection<S: DepotState + Clone + Send + Sync + 'static>(
    depot: State<S>,
    headers: HeaderMap,
    Path((depot_key, model_key)): Path<(String, String)>,
) -> impl IntoResponse {
    info!("list_item_collection {} for model {}", depot_key, model_key);
    let templates = depot.get_templates();
    let registry = depot.get_registry();
    let context = if let Ok(repo) = registry.get_repository(&depot_key, &model_key) {
        let repo = repo.lock().await;
        let depot_model = registry
            .get_model(&depot_key, &model_key)
            .expect("depot Model not found?"); // we will need a proper error route; so something that implements IntoResponse and can be substituted in the unwraps and expects.
        DepotContext {
            base: base_template(&headers),
            sections: registry.get_sections(),
            item_info: Some(repo.info(&depot_model).await),
            item_list: repo.list(&depot_model).await,
            item_model: Some(depot_model),
            ..Default::default()
        }
    } else {
        DepotContext {
            base: base_template(&headers),
            sections: registry.get_sections(),
            ..Default::default()
        }
    };
    templates.render_html("depot/items/item_list.jinja", context)
}

// Items Action is a POST to an item list. By default these are actions, that work on a list of items as input.
pub async fn item_collection_action<S: DepotState + Clone + Send + Sync + 'static>(
    depot: State<S>,
    Path((depot_key, model_key)): Path<(String, String)>,
) -> impl IntoResponse {
    "There is your answer!".to_owned()
}

// Item Details shows one single dataset.
pub async fn view_item_details<S: DepotState + Clone + Send + Sync + 'static>(
    depot: State<S>,
    headers: HeaderMap,
    Path((depot_key, model_key, id)): Path<(String, String, String)>,
) -> impl IntoResponse {
    let templates = depot.get_templates();
    let registry = depot.get_registry();
    let context = if let Ok(repo) = registry.get_repository(&depot_key, &model_key) {
        let repo = repo.lock().await;
        let depot_model = registry
            .get_model(&depot_key, &model_key)
            .expect("depot Model not found?");
        if let Some(key) = repo.key_from_string(id) {
            DepotContext {
                base: base_template(&headers),
                sections: registry.get_sections(),
                item_info: Some(repo.info(&depot_model).await),
                item_list: repo.list(&depot_model).await,
                item: repo.get(&depot_model, key.as_ref()).await.unwrap().into(),
                item_model: Some(depot_model),
                ..Default::default()
            }
        } else {
            DepotContext {
                base: base_template(&headers),
                sections: registry.get_sections(),
                ..Default::default()
            }
        }
    } else {
        DepotContext {
            base: base_template(&headers),
            sections: registry.get_sections(),
            ..Default::default()
        }
    };
    templates.render_html("depot/items/item_detail.jinja", context)
}

pub async fn new_item<S: DepotState + Clone + Send + Sync + 'static>(
    depot: State<S>,
    headers: HeaderMap,
    Path((depot_key, model_key)): Path<(String, String)>,
) -> impl IntoResponse {
    let templates = depot.get_templates();
    let registry = depot.get_registry();
    let context = if let Ok(repo) = registry.get_repository(&depot_key, &model_key) {
        let repo = repo.lock().await;
        let depot_model = registry
            .get_model(&depot_key, &model_key)
            .expect("depot Model not found?");
        DepotContext {
            base: base_template(&headers),
            sections: registry.get_sections(),
            item_info: Some(repo.info(&depot_model).await),
            item_list: repo.list(&depot_model).await,
            item_model: Some(depot_model),
            ..Default::default()
        }
    } else {
        DepotContext {
            base: base_template(&headers),
            sections: registry.get_sections(),
            ..Default::default()
        }
    };
    templates.render_html("depot/items/item_create.jinja", context)
}

pub async fn create_item<S: DepotState + Clone + Send + Sync + 'static>(
    depot: State<S>,
    headers: HeaderMap,
    Path((depot_key, model_key)): Path<(String, String)>,
    Form(form): Form<Value>,
) -> impl IntoResponse {
    let templates = depot.get_templates();
    let registry = depot.get_registry();
    let context = if let Ok(repo) = registry.get_repository(&depot_key, &model_key) {
        let mut repo = repo.lock().await;
        let depot_model = registry
            .get_model(&depot_key, &model_key)
            .expect("Depot Model not found?");

        // create our item.
        let result = repo.create(&depot_model, form).await;

        // TODO: refactor run over these views, way too much repetition.
        DepotContext {
            base: base_template(&headers),
            sections: registry.get_sections(),
            item_info: Some(repo.info(&depot_model).await),
            item_list: repo.list(&depot_model).await,
            item_model: Some(depot_model),
            item: result.unwrap().into(),
            ..Default::default()
        }
    } else {
        DepotContext {
            base: base_template(&headers),
            sections: registry.get_sections(),
            ..Default::default()
        }
    };
    templates.render_html("depot/items/item_create.jinja", context)
}

/// Change is the GET version.
pub async fn change_item<S: DepotState + Clone + Send + Sync + 'static>(
    depot: State<S>,
    headers: HeaderMap,
    Path((depot_key, model_key, id)): Path<(String, String, String)>,
) -> impl IntoResponse {
    let templates = depot.get_templates();
    let registry = depot.get_registry();
    let context = if let Ok(repo) = registry.get_repository(&depot_key, &model_key) {
        let repo = repo.lock().await;
        let depot_model = registry
            .get_model(&depot_key, &model_key)
            .expect("Depot Model not found?");
        if let Some(key) = repo.key_from_string(id) {
            DepotContext {
                base: base_template(&headers),
                sections: registry.get_sections(),
                item_info: Some(repo.info(&depot_model).await),
                item_list: repo.list(&depot_model).await,
                item: repo.get(&depot_model, key.as_ref()).await.unwrap().into(),
                item_model: Some(depot_model),
                ..Default::default()
            }
        } else {
            // Repository did not have key.
            DepotContext {
                base: base_template(&headers),
                sections: registry.get_sections(),
                ..Default::default()
            }
        }
    } else {
        // Repository could not be loaded.
        DepotContext {
            base: base_template(&headers),
            sections: registry.get_sections(),
            ..Default::default()
        }
    };
    templates.render_html("depot/items/item_change.jinja", context)
}

pub async fn update_item<S: DepotState + Clone + Send + Sync + 'static>(
    depot: State<S>,
    headers: HeaderMap,
    Path((depot_key, model_key, id)): Path<(String, String, String)>,
    Form(form): Form<Value>,
) -> impl IntoResponse {
    let templates = depot.get_templates();
    let registry = depot.get_registry();
    let context = if let Ok(repo) = registry.get_repository(&depot_key, &model_key) {
        let mut repo = repo.lock().await;
        let depot_model = registry
            .get_model(&depot_key, &model_key)
            .expect("depot Model not found?");
        if let Some(key) = repo.key_from_string(id) {
            let result = repo.update(&depot_model, key.as_ref(), form).await;
            DepotContext {
                base: base_template(&headers),
                sections: registry.get_sections(),
                item_info: Some(repo.info(&depot_model).await),
                item_list: repo.list(&depot_model).await,
                item: result.unwrap().into(),
                item_model: Some(depot_model),
                ..Default::default()
            }
        } else {
            DepotContext {
                base: base_template(&headers),
                sections: registry.get_sections(),
                ..Default::default()
            }
        }
    } else {
        DepotContext {
            base: base_template(&headers),
            sections: registry.get_sections(),
            ..Default::default()
        }
    };
    let response = templates.render_html("depot/items/item_change.jinja", context);
    response
}

// Item Action allows running an action on one single dataset.
pub async fn item_action<S: DepotState + Clone + Send + Sync + 'static>(
    depot: State<S>,
    Path((depot_key, model_key, model_id)): Path<(String, String, String)>,
) -> impl IntoResponse {
    "There is your answer!".to_owned()
}

pub async fn debug_view<S: DepotState + Clone + Send + Sync + 'static>(
    depot: State<S>,
    Path(data): Path<String>,
) -> impl IntoResponse {
    println!("debug: {}", data);
    "Debug!".to_owned()
}
